import { checkRule, hideStack } from './lib'

/**
  * SyncReplaceable function receives the whole string and returns the result of transforms which are of interface `{ re: /reg-(exp)/, replacement(m, t) { return `transform-${t}` } }`
  * If the replacement is passed as a function, it will work in the same way as the replacer for `string.replace` method (https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/replace), taking the `match` as the first argument, and matched `p1`, `p2`, _etc_ parameters as following arguments. The replacer can also be an async function.
  * @param {string|Buffer} input The string or buffer to transform synchronously using the replacements. Does not support asynchronous replacements.
  * @param {Rule[]} rules An array with rules.
  * @example
  *
  * // markdown __ to html emphasise implementation
  * const stream = SyncReplaceable('__hello world__', {
  *  --re: /__(\S+)__/g,
  *  --replacement(match, p1) {
  *  ----return `<em>${p1}</em>`
  *  --},
  * })
  */
const SyncReplaceable = (input, rules) => {
  const fr = rules.filter(checkRule)
  const s = fr.reduce((acc, { re, replacement }) => {
    /** @type {string} */
    let Acc = acc
    if (this._broke) return Acc

    if (typeof replacement == 'string') {
      Acc = Acc.replace(re, replacement)
    } else {
      /** @type {function} */
      const R = replacement.bind(this)
      let commonError
      const t = Acc.replace(re, (match, ...args) => {
        commonError = new Error()
        try {
          if (this._broke) return match
          const p = R(match, ...args)
          return p
        } catch (e) { // hide stack for sync stack traces
          hideStack(commonError, e)
        }
      })
      return t
    }
  }, `${input}`)
  return s
}

export default SyncReplaceable

/* documentary types/rules.xml */
/**
 * @typedef {(match: string, ...params: string[]) => string} Replacer
 *
 * @typedef {(match: string, ...params: string[]) => Promise.<string>} AsyncReplacer
 *
 * @typedef {Object} Rule A replacement rule.
 * @prop {RegExp} re A Regular expression to match against.
 * @prop {string|Replacer|AsyncReplacer} replacement A replacement string, or a replacement string function.
 */
