import { Transform } from 'stream'

/**
 * Create a Transform stream which will maintain a buffer with data received from a Readable stream and write data when the buffer can be matched against the regex. It will push the whole match object (or objects when the g flag is used) returned by `/regex/.exec(buffer)`.
 * @param {RegExp} regex Regular expression to execute
 */
export default function createRegexTransformStream(regex) {
  let buffer = ''

  const ts = new Transform({
    transform(chunk, encoding, next) {
      let lastMatch
      let match
      buffer += chunk.toString()

      // thx stream-snitch
      // https://github.com/dmotz/stream-snitch/blob/master/index.js#L52
      // eslint-disable-next-line no-cond-assign
      while (match = regex.exec(buffer)) {
        ts.push(match)
        lastMatch = match
        if (!regex.global) break
      }
      if (lastMatch) {
        buffer = buffer.slice(lastMatch.index + lastMatch[0].length)
      }
      next()
    },
    objectMode: true,
  })

  return ts
}

export {
  makeMarkers,
  makeCutRule,
  makePasteRule,
} from './lib/markers'

export { default as Replaceable, SerialAsyncReplaceable } from './Replaceable'
export { default as SyncReplaceable } from './SyncReplaceable'

/* documentary types/rules.xml */
/**
 * @typedef {(match: string, ...params: string[]) => string} Replacer
 *
 * @typedef {(match: string, ...params: string[]) => Promise.<string>} AsyncReplacer
 *
 * @typedef {Object} Rule A replacement rule.
 * @prop {RegExp} re A Regular expression to match against.
 * @prop {string|Replacer|AsyncReplacer} replacement A replacement string, or a replacement string function.
 */
